from pitop.core.mixins import Recreatable, Stateful
from pitop.pma import ServoMotor
from pitop.robotics.two_servo_assembly_calibrator import TwoServoAssemblyCalibrator

from .pan_tilt_object_tracker import PanTiltObjectTracker


class PanTiltController(Stateful, Recreatable):
    CALIBRATION_FILE_NAME = "pan_tilt.conf"
    _pan_servo = None
    _tilt_servo = None

    def __init__(self, servo_pan_port="S0", servo_tilt_port="S3", name="pan_tilt"):
        self.name = name
        self._pan_servo = ServoMotor(servo_pan_port)
        self._tilt_servo = ServoMotor(servo_tilt_port)

        self._object_tracker = PanTiltObjectTracker(
            pan_servo=self._pan_servo, tilt_servo=self._tilt_servo
        )

        Stateful.__init__(self, children=["_pan_servo", "_tilt_servo"])
        Recreatable.__init__(
            self,
            config_dict={
                "servo_pan_port": servo_pan_port,
                "servo_tilt_port": servo_tilt_port,
                "name": name,
            },
        )

    @property
    def pan_servo(self):
        return self._pan_servo

    @property
    def tilt_servo(self):
        return self._tilt_servo

    @property
    def track_object(self) -> PanTiltObjectTracker:
        return self._object_tracker

    def calibrate(self, save=True, reset=False):
        """Calibrates the assembly to work in optimal conditions.

        Based on the provided arguments, it will either load the calibration
        values stored in the pi-top, or it will run the calibration process,
        requesting the user input in an interactive fashion.

        :param bool reset:
            If `true`, the existing calibration values will be reset, and the calibration process will be started.
            If set to `false`, the calibration values will be retrieved from the calibration file.
        :param bool save:
            If `reset` is `true`, this parameter will cause the calibration values to be stored to the calibration file if set to `true`.
            If `save=False`, the calibration values will only be used for the current session.
        """
        calibration_object = TwoServoAssemblyCalibrator(
            filename=self.CALIBRATION_FILE_NAME,
            section_name="PAN_TILT",
            servo_lookup_dict={
                "pan_zero_point": self.pan_servo,
                "tilt_zero_point": self.tilt_servo,
            },
        )
        calibration_object.calibrate(save, reset)



import time
from dataclasses import dataclass
from threading import Thread
from time import sleep
from typing import Optional

from pitop.core.mixins import Recreatable, Stateful
from pitop.pma import ServoMotor
from pitop.robotics.two_servo_assembly_calibrator import TwoServoAssemblyCalibrator

from .simple_pid import PID


@dataclass
class OscillateRequest:
    servo: ServoMotor
    times: int
    angle: int
    speed: int
    block: bool


class ThreadControl:
    def __init__(self):
        self.thread = Thread()
        self.cancel = False


class TiltRollHeadController(Stateful, Recreatable):
    CALIBRATION_FILE_NAME = "tilt_roll.conf"
    _roll_servo = None
    _tilt_servo = None

    def __init__(self, servo_roll_port="S0", servo_tilt_port="S3", name="head"):
        self.name = name
        self._roll_servo = ServoMotor(servo_roll_port)
        self._tilt_servo = ServoMotor(servo_tilt_port)

        self._head_roll_pid = PID(
            Kp=3.0, Ki=1.5, Kd=0.25, setpoint=0, output_limits=(-100, 100)
        )

        self.timeout = 5  # s
        self.__shake_thread_control = ThreadControl()
        self.__nod_thread_control = ThreadControl()

        Stateful.__init__(self, children=["_roll_servo", "_tilt_servo"])
        Recreatable.__init__(
            self,
            config_dict={
                "servo_roll_port": servo_roll_port,
                "servo_tilt_port": servo_tilt_port,
                "name": name,
            },
        )

    @property
    def roll(self) -> ServoMotor:
        return self._roll_servo

    @property
    def tilt(self) -> ServoMotor:
        return self._tilt_servo

    def stop(self) -> None:
        self.tilt.sweep(speed=0)
        self.roll.sweep(speed=0)

    def shake(
        self, times: int = 4, angle: int = 5, speed: int = 100, block: bool = True
    ) -> None:
        """Shakes the head by rotating the roll servo back and forth between
        :data:`-angle` and :data:`+angle` from current angle position.

        :param int times: Number of times to shake head
        :param int angle: angle in degrees either side of current angle to rotate servo by
        :param int speed: servo speed in deg/s
        :param bool block: if `True`, function call will block program execution until finished
        :return: None
        """
        oscillate_request = OscillateRequest(
            servo=self._roll_servo, times=times, angle=angle, speed=speed, block=block
        )

        self.__start_servo_oscillation(
            oscillate_request, thread_control=self.__shake_thread_control
        )

    def nod(
        self, times: int = 4, angle: int = 5, speed: int = 100, block: bool = True
    ) -> None:
        """Nod the head by rotating the tilt servo back and forth between
        :data:`-angle` and :data:`+angle` from current angle position.

        :param int times: Number of times to nod head
        :param int angle: angle in degrees either side of current angle to rotate servo by
        :param int speed: servo speed in deg/s
        :param bool block: if `True`, function call will block program execution until finished
        :return: None
        """
        oscillate_request = OscillateRequest(
            servo=self._tilt_servo, times=times, angle=angle, speed=speed, block=block
        )

        self.__start_servo_oscillation(
            oscillate_request, thread_control=self.__nod_thread_control
        )

    def __start_servo_oscillation(
        self, oscillate_request: OscillateRequest, thread_control: ThreadControl
    ):
        if thread_control.thread.is_alive():
            thread_control.cancel = True
            thread_control.thread.join()
            thread_control.cancel = False

        if oscillate_request.block:
            self.__oscillate_servo(
                oscillate_request=oscillate_request, thread_control=thread_control
            )
        else:
            thread_control.thread = Thread(
                target=self.__oscillate_servo,
                kwargs={
                    "oscillate_request": oscillate_request,
                    "thread_control": thread_control,
                },
                daemon=True,
            )
            thread_control.thread.start()

    def __oscillate_servo(
        self, oscillate_request: OscillateRequest, thread_control: ThreadControl
    ) -> None:
        def reset_servo_state():
            oscillate_request.servo.target_speed = previous_target_speed
            self.__set_angle_until_reached(
                servo=oscillate_request.servo, angle=starting_angle
            )

        previous_target_speed = oscillate_request.servo.target_speed
        oscillate_request.servo.target_speed = oscillate_request.speed

        starting_angle = oscillate_request.servo.current_angle
        angle_start = starting_angle - oscillate_request.angle
        angle_end = starting_angle + oscillate_request.angle

        for _ in range(oscillate_request.times):
            if thread_control.cancel:
                reset_servo_state()
                return

            self.__set_angle_until_reached(
                servo=oscillate_request.servo,
                angle=angle_start,
                thread_control=thread_control,
            )
            self.__set_angle_until_reached(
                servo=oscillate_request.servo,
                angle=angle_end,
                thread_control=thread_control,
            )

        reset_servo_state()

    def __set_angle_until_reached(
        self,
        servo: ServoMotor,
        angle: int,
        thread_control: Optional[ThreadControl] = None,
    ):
        def cancelled():
            if time.time() - start_time > self.timeout:
                servo.stop()
                thread_control.cancel = True
                print(
                    f"Head nod/shake cancelled because time exceeded {self.timeout} seconds. Try a higher speed."
                )
                return True
            if thread_control is not None:
                return thread_control.cancel
            return False

        start_time = time.time()
        if not cancelled():
            servo.target_angle = angle

            while abs(angle - servo.current_angle) > 1:
                if cancelled():
                    break
                sleep(0.05)

    def track_head_angle(self, angle: int, flipped=True) -> None:
        if not flipped:
            angle = -angle
        current_angle = self.roll.current_angle
        error = current_angle - angle
        if abs(error) < 1.0:
            self.roll.sweep(speed=0)
        else:
            servo_speed = self._head_roll_pid(error)
            self.roll.sweep(speed=servo_speed)

    def calibrate(self, save: bool = True, reset: bool = False) -> None:
        """Calibrates the assembly to work in optimal conditions.

        Based on the provided arguments, it will either load the calibration
        values stored in the pi-top, or it will run the calibration process,
        requesting the user input in an interactive fashion.

        :param bool reset:
            If `true`, the existing calibration values will be reset, and the calibration process will be started.
            If set to `false`, the calibration values will be retrieved from the calibration file.
        :param bool save:
            If `reset` is `true`, this parameter will cause the calibration values to be stored to the calibration file if set to `true`.
            If `save=False`, the calibration values will only be used for the current session.
        """
        calibration_object = TwoServoAssemblyCalibrator(
            filename=self.CALIBRATION_FILE_NAME,
            section_name="TILT_ROLL",
            servo_lookup_dict={
                "roll_zero_point": self.roll,
                "tilt_zero_point": self.tilt,
            },
        )
        calibration_object.calibrate(save, reset)



from math import floor, radians
from time import sleep

from pitop.core.mixins import Recreatable, Stateful
from pitop.pma import EncoderMotor, ForwardDirection
from pitop.pma.common.encoder_motor_registers import MotorSyncBits, MotorSyncRegisters
from pitop.pma.plate_interface import PlateInterface

from .simple_pid import PID


class DriveController(Stateful, Recreatable):
    """Represents a vehicle with two wheels connected by an axis, and an
    optional support wheel or caster."""

    def __init__(self, left_motor_port="M3", right_motor_port="M0", name="drive"):
        self.name = name

        # motor and wheel setup
        self.left_motor_port = left_motor_port
        self.right_motor_port = right_motor_port
        self.left_motor = EncoderMotor(
            port_name=self.left_motor_port, forward_direction=ForwardDirection.CLOCKWISE
        )
        self.right_motor = EncoderMotor(
            port_name=self.right_motor_port,
            forward_direction=ForwardDirection.COUNTER_CLOCKWISE,
        )

        # chassis setup
        self.wheel_separation = 0.163

        # Round down to ensure no speed value ever goes above maximum due to rounding issues (resulting in error)
        self.max_motor_speed = (
            floor(min(self.left_motor.max_speed, self.right_motor.max_speed) * 1000)
            / 1000
        )
        self.max_robot_angular_speed = self.max_motor_speed / (
            self.wheel_separation / 2
        )

        # Target lock drive angle
        self._linear_speed_x_hold = 0
        self.__target_lock_pid_controller = PID(
            Kp=0.045,
            Ki=0.002,
            Kd=0.0035,
            setpoint=0,
            output_limits=(-self.max_robot_angular_speed, self.max_robot_angular_speed),
        )

        # Motor syncing
        self.__mcu_device = PlateInterface().get_device_mcu()
        self._set_synchronous_motor_movement_mode()

        Stateful.__init__(self, children=["left_motor", "right_motor"])
        Recreatable.__init__(
            self,
            config_dict={
                "left_motor_port": left_motor_port,
                "right_motor_port": right_motor_port,
                "name": self.name,
            },
        )

    def _set_synchronous_motor_movement_mode(self):
        sync_config = (
            MotorSyncBits[self.left_motor_port].value
            | MotorSyncBits[self.right_motor_port].value
        )
        self.__mcu_device.write_byte(MotorSyncRegisters.CONFIG.value, sync_config)

    def _start_synchronous_motor_movement(self):
        self.__mcu_device.write_byte(MotorSyncRegisters.START.value, 1)

    def _calculate_motor_speeds(self, linear_speed, angular_speed, turn_radius):
        # if angular_speed is positive, then rotation is anti-clockwise in this coordinate frame
        speed_right = (
            linear_speed + (turn_radius + self.wheel_separation / 2) * angular_speed
        )
        speed_left = (
            linear_speed + (turn_radius - self.wheel_separation / 2) * angular_speed
        )

        if (
            abs(speed_right) > self.max_motor_speed
            or abs(speed_left) > self.max_motor_speed
        ):
            factor = self.max_motor_speed / max(abs(speed_left), abs(speed_right))
            speed_right = speed_right * factor
            speed_left = speed_left * factor

        return speed_left, speed_right

    def robot_move(self, linear_speed, angular_speed, turn_radius=0.0):
        # TODO: turn_radius will introduce a hidden linear speed component to the robot, so params are syntactically
        #  misleading
        speed_left, speed_right = self._calculate_motor_speeds(
            linear_speed, angular_speed, turn_radius
        )
        self.left_motor.set_target_speed(target_speed=speed_left)
        self.right_motor.set_target_speed(target_speed=speed_right)
        self._start_synchronous_motor_movement()

    def forward(self, speed_factor, hold=False):
        """Move the robot forward.

        :param float speed_factor: Factor relative to the maximum motor
            speed, used to set the velocity, in the range -1.0 to 1.0.
            Using negative values will cause the robot to move
            backwards.
        :param bool hold: Setting this parameter to true will cause
            subsequent movements to use the speed set as the base speed.
        """
        linear_speed_x = self.max_motor_speed * speed_factor
        if hold:
            self._linear_speed_x_hold = linear_speed_x
        else:
            self._linear_speed_x_hold = 0
        self.robot_move(linear_speed_x, 0)

    def backward(self, speed_factor, hold=False):
        """Move the robot backward.

        :param float speed_factor: Factor relative to the maximum motor
            speed, used to set the velocity, in the range -1.0 to 1.0.
            Using negative values will cause the robot to move forwards.
        :param bool hold: Setting this parameter to true will cause
            subsequent movements to use the speed set as the base speed.
        """
        self.forward(-speed_factor, hold)

    def left(self, speed_factor, turn_radius=0):
        """Make the robot move to the left, using a circular trajectory.

        :param float speed_factor: Factor relative to the maximum motor
            speed, used to set the velocity, in the range -1.0 to 1.0.
            Using negative values will cause the robot to turn right.
        :param float turn_radius: Radius used by the robot to perform
            the movement. Using `turn_radius=0` will cause the robot to
            rotate in place.
        """

        self.robot_move(
            self._linear_speed_x_hold,
            self.max_robot_angular_speed * speed_factor,
            turn_radius,
        )

    def right(self, speed_factor, turn_radius=0):
        """Make the robot move to the right, using a circular trajectory.

        :param float speed_factor: Factor relative to the maximum motor
            speed, used to set the velocity, in the range -1.0 to 1.0.
            Using negative values will cause the robot to turn left.
        :param float turn_radius: Radius used by the robot to perform
            the movement. Using `turn_radius=0` will cause the robot to
            rotate in place.
        """

        self.left(-speed_factor, -turn_radius)

    def target_lock_drive_angle(self, angle):
        """Make the robot move in the direction of the specified angle, while
        maintaining the current linear speed."""
        angular_speed = self.__target_lock_pid_controller(angle)
        self.robot_move(self._linear_speed_x_hold, angular_speed)

    def rotate(self, angle, time_to_take):
        """Rotate the robot in place by a given angle and stop.

        :param float angle: Angle of the turn.
        :param float time_to_take: Expected duration of the rotation, in
            seconds.
        """
        assert time_to_take > 0.0
        angle_radians = radians(angle)
        angular_speed = angle_radians / time_to_take

        speed_left, speed_right = self._calculate_motor_speeds(
            0, angular_speed, turn_radius=0
        )
        distance = abs(angle_radians) * self.wheel_separation / 2
        self.left_motor.set_target_speed(
            target_speed=speed_left, distance=distance * speed_left / abs(speed_left)
        )
        self.right_motor.set_target_speed(
            target_speed=speed_right, distance=distance * speed_right / abs(speed_right)
        )
        sleep(time_to_take)

    def stop(self):
        """Stop any movement being performed by the motors."""
        self._linear_speed_x_hold = 0
        self.robot_move(0, 0)

    def stop_rotation(self):
        """Stop any angular movement performed by the robot.

        In the case where linear and rotational movements are being
        performed at the same time (e.g.: during a left turn with a turn
        radius different to 0), calling this method will cause the robot
        to continue the linear movement, so it will continue to move
        forward.
        """
        self.robot_move(self._linear_speed_x_hold, 0)



import sched
import time
from threading import Event, Thread


class MeasurementScheduler:
    def __init__(
        self, measurement_frequency, measurement_input_function, state_tracker
    ):
        if not callable(measurement_input_function):
            raise AttributeError(
                "Argument 'measurement_input_function' must be a function"
            )

        self.measurement_func = measurement_input_function
        self.state_tracker = state_tracker
        self._measurement_dt = 1.0 / measurement_frequency

        self._continue_processing = True
        self._new_measurement_event = Event()
        self._measurement_prediction_scheduler = Thread(target=self.start, daemon=True)
        self._measurement_prediction_scheduler.start()

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self._continue_processing = False
        if self._measurement_prediction_scheduler.is_alive():
            self._measurement_prediction_scheduler.join()

    def wait_for_measurement(self):
        self._new_measurement_event.wait()

    def start(self):
        s = sched.scheduler(time.time, time.sleep)
        current_time = time.time()
        s.enterabs(
            current_time + self._measurement_dt,
            1,
            self.loop,
            (s, current_time),
        )
        s.run()

    def loop(self, s, previous_time):
        current_time = time.time()
        self.state_tracker.add_measurements(
            odom_measurements=self.measurement_func(), dt=current_time - previous_time
        )

        self._new_measurement_event.set()
        self._new_measurement_event.clear()

        if self._continue_processing:
            s.enterabs(
                current_time + self._measurement_dt,
                1,
                self.loop,
                (s, current_time),
            )



from .blockpi_rover import BlockPiRover
from .configurations import AlexRobot  # deprecated
from .configurations import alex_config
from .drive_controller import DriveController
from .navigation import NavigationController
from .pan_tilt_controller import PanTiltController
from .pincer_controller import PincerController
from .tilt_roll_head_controller import TiltRollHeadController



from threading import Event, Thread
from typing import Union

import numpy as np

from ..drive_controller import DriveController
from .core.measurement_scheduler import MeasurementScheduler
from .core.robot_state import StateFilter
from .core.utils import verify_callback
from .navigator import Navigator


class NavigationController(DriveController):
    __VALID_ANGLE_RANGE = 180

    def __init__(
        self,
        left_motor_port: str = "M3",
        right_motor_port: str = "M0",
        linear_speed_factor: float = 0.75,
        angular_speed_factor: float = 0.5,
        name="navigate",
    ):
        """
        :param left_motor_port: string with the name of the port for the left motor
        :param right_motor_port: string with the name of the port for the right motor
        :param linear_speed_factor: value between 0 and 1 for the linear speed
        :param angular_speed_factor: value between 0 and 1 for the angular speed
        """
        super().__init__(
            left_motor_port=left_motor_port,
            right_motor_port=right_motor_port,
            name=name,
        )

        # callback to call once navigation complete
        self._on_finish = None

        # Navigation algorithm flow control
        self.in_progress = False
        self._nav_goal_finish_event = Event()
        self._nav_thread = None

        self._measurement_frequency = 10.0  # Hz

        # Robot state tracking and driving management
        self.state_tracker = StateFilter(
            measurement_frequency=self._measurement_frequency,
            wheel_separation=self.wheel_separation,
        )

        self.navigator = Navigator(
            max_motor_speed=self.max_motor_speed,
            max_robot_angular_speed=self.max_robot_angular_speed,
            measurement_input_function=lambda: self.pose,
        )

        # Odometry updates
        self.measurement_scheduler = MeasurementScheduler(
            measurement_frequency=self._measurement_frequency,
            measurement_input_function=lambda: self.odometry,
            state_tracker=self.state_tracker,
        )

        self.linear_speed_factor = linear_speed_factor
        self.angular_speed_factor = angular_speed_factor

    @property
    def own_state(self):
        return {
            "in_progress": self.in_progress,
            "tracker": self.state_tracker.own_state,
        }

    def go_to(
        self,
        position: Union[tuple, None] = None,
        angle: Union[float, int, None] = None,
        on_finish=None,
        backwards: bool = False,
    ):
        """Navigates the robot to a position (x, y) and/or angle where the
        starting position is assumed to be (0, 0) and the starting angle is
        assumed to be 0 degrees. Call function with .wait() appended to block
        program execution until the navigation goal has been achieved.

        Calling this function whilst another navigation goal is in
        progress will raise an error. Use the .wait() function to wait
        until navigation is completed or call .stop() if you wish to set
        a new navigation goal.

        :param Union[tuple, None] position: Position goal tuple in the
        form (x, y) where x and y are in meters (int or float). :param
        Union[float, int, None] angle: Desired angle of the robot in
        degrees.
        :param on_finish: A callable function or class to be called when
            the navigation goal has been reached.
        :param bool backwards: Go to navigation goal in reverse by
            setting to True
        :return NavigationController: self
        """
        if self.in_progress:
            raise RuntimeError(
                "Cannot call function before previous navigation is complete, use .wait() or call "
                ".stop() to cancel the previous navigation request."
            )

        if position is not None:
            if len(position) != 2 or type(position) not in (tuple, list):
                raise ValueError(
                    f"Position should be a list of size two in the form [x, y]. Received {position}."
                )
            if not all(isinstance(coordinate, (int, float)) for coordinate in position):
                raise ValueError("x and y coordinates must be of type int or float.")

        if angle is not None:
            if not (-self.__VALID_ANGLE_RANGE <= angle <= self.__VALID_ANGLE_RANGE):
                raise ValueError(
                    f"Angle must from {-self.__VALID_ANGLE_RANGE} to {self.__VALID_ANGLE_RANGE}."
                )

        self._on_finish = verify_callback(on_finish)
        self._nav_thread = Thread(
            target=self.__navigate,
            args=(
                position,
                angle,
                backwards,
            ),
            daemon=True,
        )
        self._nav_thread.start()

        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.stop()

    def __navigate(self, position, angle, backwards):
        self.__navigation_started()

        self.navigator.backwards = backwards
        for linear_speed, angular_speed in self.navigator.navigate(position, angle):
            self.robot_move(
                linear_speed=linear_speed,
                angular_speed=angular_speed,
            )

        self.__navigation_finished()

    def wait(self, timeout: Union[float, None] = None):
        """Call this to pause your program execution until the navigation
        request is complete."""
        self._nav_goal_finish_event.wait(timeout)

    @property
    def pose(self):
        self.measurement_scheduler.wait_for_measurement()
        return self.state_tracker.x, self.state_tracker.y, self.state_tracker.angle_rad

    @property
    def odometry(self):
        left_wheel_speed = self.left_motor.current_speed
        right_wheel_speed = self.right_motor.current_speed
        linear_velocity = (right_wheel_speed + left_wheel_speed) / 2.0
        angular_velocity = (
            right_wheel_speed - left_wheel_speed
        ) / self.wheel_separation

        return np.array([[linear_velocity], [angular_velocity]])

    @property
    def linear_speed_factor(self):
        """
        :return: current linear speed factor
        """
        return self.navigator.linear_speed_factor

    @linear_speed_factor.setter
    def linear_speed_factor(self, speed_factor: float):
        """Update the linear speed factor to change the speed that the robot
        will try to reach the goal position. Increasing this value will
        increase the uncertainty in the final navigation position. Setting to
        zero is not permitted.

        :param float speed_factor: Value greater than 0.0 and less than
            or equal to 1.0 where 1.0 is the maximum linear speed of the
            robot (which is based on maximum motor RPM and wheel
            circumference).
        """
        self.navigator.linear_speed_factor = speed_factor

    @property
    def angular_speed_factor(self):
        """
        :return: current angular speed factor
        """
        return self.navigator.angular_speed_factor

    @angular_speed_factor.setter
    def angular_speed_factor(self, speed_factor):
        """Update the angular speed factor to change the speed that the robot
        will try to reach any angle goals. Increasing this value will increase
        the uncertainty in the final navigation position and angle. Setting to
        zero is not permitted.

        :param float speed_factor: Value greater than 0.0 and less than
            or equal to 1.0 where 1.0 is the maximum angular speed of
            the robot (which is based on maximum motor RPM, wheel
            circumference and wheel-to-wheel spacing).
        """
        self.navigator.angular_speed_factor = speed_factor

    def reset_position_and_angle(self):
        """Reset the robot's position and angle (pose) to zeros."""
        self.state_tracker.reset_pose()

    def stop(self):
        """Terminate the navigation goal that is currently in progress and stop
        the robot's movement.

        Any on_finish callback function passed to the go_to() method
        will not be called if this function is called before the
        navigation goal has been reached.
        """
        self.stop_navigation()
        self.stop_movement()

    def stop_navigation(self):
        # don't call callback if user has terminated navigation manually
        self._on_finish = None
        self.navigator._stop_triggered = True
        try:
            self._nav_thread.join()
        except Exception:
            raise
        self.__navigation_finished()

    def stop_movement(self):
        super().stop()

    def __navigation_started(self):
        self.in_progress = True
        self.navigator._stop_triggered = False

    def __navigation_finished(self):
        self.in_progress = False
        self._nav_goal_finish_event.set()
        self._nav_goal_finish_event.clear()
        if callable(self._on_finish):
            self._on_finish()



import warnings

import numpy as np
from scipy.stats import multivariate_normal

# Older versions of scipy do not support the allow_singular keyword. I could
# check the version number explicily, but perhaps this is clearer
_support_singular = True
try:
    multivariate_normal.logpdf(1, 1, 1, allow_singular=True)
except TypeError:
    warnings.warn(
        "You are using a version of SciPy that does not support the "
        "allow_singular parameter in scipy.stats.multivariate_normal.logpdf(). "
        "Future versions of FilterPy will require a version of SciPy that "
        "implements this keyword",
        DeprecationWarning,
    )
    _support_singular = False


def logpdf(x, mean=None, cov=1, allow_singular=True):
    """Computes the log of the probability density function of the normal
    N(mean, cov) for the data x.

    The normal may be univariate or multivariate. Wrapper for older
    versions of scipy.multivariate_normal.logpdf which don't support
    support the allow_singular keyword prior to verion 0.15.0. If it is
    not supported, and cov is singular or not PSD you may get an
    exception. `x` and `mean` may be column vectors, row vectors, or
    lists.
    """

    if mean is not None:
        flat_mean = np.asarray(mean).flatten()
    else:
        flat_mean = None

    flat_x = np.asarray(x).flatten()

    if _support_singular:
        return multivariate_normal.logpdf(flat_x, flat_mean, cov, allow_singular)
    return multivariate_normal.logpdf(flat_x, flat_mean, cov)



import math


def normalize_angle(angle):
    """Converts to range -pi to +pi to prevent unstable behaviour when going
    from 0 to 2*pi with slight turn.

    :param angle: angle in radians
    :return: angle in radians normalized to range -pi to +pi
    """
    return (angle + math.pi) % (2 * math.pi) - math.pi


def verify_callback(callback):
    if callback is None:
        return None
    if not callable(callback):
        raise ValueError("callback should be a callable function.")

    from inspect import getfullargspec

    arg_spec = getfullargspec(callback)
    number_of_arguments = len(arg_spec.args)
    number_of_default_arguments = (
        len(arg_spec.defaults) if arg_spec.defaults is not None else 0
    )
    if number_of_arguments == 0:
        return callback
    if (
        arg_spec.args[0] in ("self", "_mock_self")
        and (number_of_arguments - number_of_default_arguments) == 1
    ):
        return callback
    if number_of_arguments != number_of_default_arguments:
        raise ValueError("callback should have no non-default keyword arguments.")
    return callback



import numpy as np


def pretty_str(label, arr):
    """Generates a pretty printed NumPy array with an assignment. Optionally
    transposes column vectors so they are drawn on one line. Strictly speaking
    arr can be any time convertible by `str(arr)`, but the output may not be
    what you want if the type of the variable is not a scalar or an ndarray.

    Examples
    --------
    >>> pprint('cov', np.array([[4., .1], [.1, 5]]))
    cov = [[4.  0.1]
           [0.1 5. ]]
    >>> print(pretty_str('x', np.array([[1], [2], [3]])))
    x = [[1 2 3]].T
    """

    def is_col(a):
        """Return true if a is a column vector."""
        try:
            return a.shape[0] > 1 and a.shape[1] == 1
        except (AttributeError, IndexError):
            return False

    if label is None:
        label = ""

    if label:
        label += " = "

    if is_col(arr):
        return label + str(arr.T).replace("\n", "") + ".T"

    rows = str(arr).split("\n")
    if not rows:
        return ""

    s = label + rows[0]
    pad = " " * len(label)
    for line in rows[1:]:
        s = s + "\n" + pad + line

    return s


def reshape_z(z, dim_z, ndim):
    """Ensure z is a (dim_z, 1) shaped vector."""

    z = np.atleast_2d(z)
    if z.shape[1] == dim_z:
        z = z.T

    if z.shape != (dim_z, 1):
        raise ValueError(
            "z (shape {}) must be convertible to shape ({}, 1)".format(z.shape, dim_z)
        )

    if ndim == 1:
        z = z[:, 0]

    if ndim == 0:
        z = z[0, 0]

    return z



import time
import warnings


def _clamp(value, limits):
    lower, upper = limits
    if value is None:
        return None
    elif (upper is not None) and (value > upper):
        return upper
    elif (lower is not None) and (value < lower):
        return lower
    return value


try:
    # Get monotonic time to ensure that time deltas are always positive
    _current_time = time.monotonic
except AttributeError:
    # time.monotonic() not available (using python < 3.3), fallback to time.time()
    _current_time = time.time
    warnings.warn(
        "time.monotonic() not available in python < 3.3, using time.time() as fallback"
    )


class PID(object):
    """A simple PID controller."""

    def __init__(
        self,
        Kp=1.0,
        Ki=0.0,
        Kd=0.0,
        setpoint=0,
        sample_time=0.01,
        output_limits=(None, None),
        auto_mode=True,
        proportional_on_measurement=False,
        error_map=None,
    ):
        """Initialize a new PID controller.

        :param Kp: The value for the proportional gain Kp
        :param Ki: The value for the integral gain Ki
        :param Kd: The value for the derivative gain Kd
        :param setpoint: The initial setpoint that the PID will try to
            achieve
        :param sample_time: The time in seconds which the controller
            should wait before generating a new output value. The PID
            works best when it is constantly called (eg. during a loop),
            but with a sample time set so that the time difference
            between each update is (close to) constant. If set to None,
            the PID will compute a new output value every time it is
            called.
        :param output_limits: The initial output limits to use, given as
            an iterable with 2 elements, for example: (lower, upper).
            The output will never go below the lower limit or above the
            upper limit. Either of the limits can also be set to None to
            have no limit in that direction. Setting output limits also
            avoids integral windup, since the integral term will never
            be allowed to grow outside of the limits.
        :param auto_mode: Whether the controller should be enabled (auto
            mode) or not (manual mode)
        :param proportional_on_measurement: Whether the proportional
            term should be calculated on the input directly rather than
            on the error (which is the traditional way). Using
            proportional-on-measurement avoids overshoot for some types
            of systems.
        :param error_map: Function to transform the error value in
            another constrained value.
        """
        self.Kp, self.Ki, self.Kd = Kp, Ki, Kd
        self.setpoint = setpoint
        self.sample_time = sample_time

        self._min_output, self._max_output = None, None
        self._auto_mode = auto_mode
        self.proportional_on_measurement = proportional_on_measurement
        self.error_map = error_map

        self._proportional = 0
        self._integral = 0
        self._derivative = 0

        self._last_time = None
        self._last_output = None
        self._last_input = None

        self.output_limits = output_limits
        self.reset()

    def __call__(self, input_, dt=None):
        """Update the PID controller.

        Call the PID controller with *input_* and calculate and return a
        control output if sample_time seconds has passed since the last
        update. If no new output is calculated, return the previous
        output instead (or None if no value has been calculated yet).

        :param dt: If set, uses this value for timestep instead of real
            time. This can be used in simulations when simulation time
            is different from real time.
        """
        if not self.auto_mode:
            return self._last_output

        now = _current_time()
        if dt is None:
            dt = now - self._last_time if (now - self._last_time) else 1e-16
        elif dt <= 0:
            raise ValueError("dt has negative value {}, must be positive".format(dt))

        if (
            self.sample_time is not None
            and dt < self.sample_time
            and self._last_output is not None
        ):
            # Only update every sample_time seconds
            return self._last_output

        # Compute error terms
        error = self.setpoint - input_
        d_input = input_ - (
            self._last_input if (self._last_input is not None) else input_
        )

        # Check if must map the error
        if self.error_map is not None:
            error = self.error_map(error)

        # Compute the proportional term
        if not self.proportional_on_measurement:
            # Regular proportional-on-error, simply set the proportional term
            self._proportional = self.Kp * error
        else:
            # Add the proportional error on measurement to error_sum
            self._proportional -= self.Kp * d_input

        # Compute integral and derivative terms
        self._integral += self.Ki * error * dt
        self._integral = _clamp(
            self._integral, self.output_limits
        )  # Avoid integral windup

        self._derivative = -self.Kd * d_input / dt

        # Compute final output
        output = self._proportional + self._integral + self._derivative
        output = _clamp(output, self.output_limits)

        # Keep track of state
        self._last_output = output
        self._last_input = input_
        self._last_time = now

        return output

    def __repr__(self):
        return (
            "{self.__class__.__name__}("
            "Kp={self.Kp!r}, Ki={self.Ki!r}, Kd={self.Kd!r}, "
            "setpoint={self.setpoint!r}, sample_time={self.sample_time!r}, "
            "output_limits={self.output_limits!r}, auto_mode={self.auto_mode!r}, "
            "proportional_on_measurement={self.proportional_on_measurement!r},"
            "error_map={self.error_map!r}"
            ")"
        ).format(self=self)

    @property
    def components(self):
        """The P-, I- and D-terms from the last computation as separate
        components as a tuple.

        Useful for visualizing what the controller is doing or when
        tuning hard-to-tune systems.
        """
        return self._proportional, self._integral, self._derivative

    @property
    def tunings(self):
        """The tunings used by the controller as a tuple: (Kp, Ki, Kd)."""
        return self.Kp, self.Ki, self.Kd

    @tunings.setter
    def tunings(self, tunings):
        """Set the PID tunings."""
        self.Kp, self.Ki, self.Kd = tunings

    @property
    def auto_mode(self):
        """Whether the controller is currently enabled (in auto mode) or
        not."""
        return self._auto_mode

    @auto_mode.setter
    def auto_mode(self, enabled):
        """Enable or disable the PID controller."""
        self.set_auto_mode(enabled)

    def set_auto_mode(self, enabled, last_output=None):
        """Enable or disable the PID controller, optionally setting the last
        output value.

        This is useful if some system has been manually controlled and
        if the PID should take over. In that case, disable the PID by
        setting auto mode to False and later when the PID should be
        turned back on, pass the last output variable (the control
        variable) and it will be set as the starting I-term when the PID
        is set to auto mode.

        :param enabled: Whether auto mode should be enabled, True or
            False
        :param last_output: The last output, or the control variable,
            that the PID should start from when going from manual mode
            to auto mode. Has no effect if the PID is already in auto
            mode.
        """
        if enabled and not self._auto_mode:
            # Switching from manual mode to auto, reset
            self.reset()

            self._integral = last_output if (last_output is not None) else 0
            self._integral = _clamp(self._integral, self.output_limits)

        self._auto_mode = enabled

    @property
    def output_limits(self):
        """The current output limits as a 2-tuple: (lower, upper).

        See also the *output_limits* parameter in :meth:`PID.__init__`.
        """
        return self._min_output, self._max_output

    @output_limits.setter
    def output_limits(self, limits):
        """Set the output limits."""
        if limits is None:
            self._min_output, self._max_output = None, None
            return

        min_output, max_output = limits

        if (None not in limits) and (max_output < min_output):
            raise ValueError("lower limit must be less than upper limit")

        self._min_output = min_output
        self._max_output = max_output

        self._integral = _clamp(self._integral, self.output_limits)
        self._last_output = _clamp(self._last_output, self.output_limits)

    def reset(self):
        """Reset the PID controller internals.

        This sets each term to 0 as well as clearing the integral, the
        last output and the last input (derivative calculation).
        """
        self._proportional = 0
        self._integral = 0
        self._derivative = 0

        self._integral = _clamp(self._integral, self.output_limits)

        self._last_time = _current_time()
        self._last_output = None
        self._last_input = None



from json import load
from pathlib import Path


def __load_json(filename):
    path = __robotics_directory() / "json" / filename
    with open(path, "r") as handle:
        config = load(handle)
    return config


def __robotics_directory() -> Path:
    return Path(__file__).parent


alex_config = __load_json("alex.json")

bobbie_config = __load_json("bobbie.json")


class AlexRobot:
    def __init__(self, *args, **kwargs):
        print(
            "AlexRobot class is deprecated. Please use Pitop.from_config(alex_config)"
        )

    def __new__(cls, *args, **kwargs):
        from pitop.system.pitop import Pitop

        obj = Pitop.from_config(alex_config)
        obj.__class__ = cls
        return obj

    def forward(self, *args, **kwargs):
        if hasattr(self, "drive"):
            return self.drive.forward(*args, **kwargs)

    def backward(self, *args, **kwargs):
        if hasattr(self, "drive"):
            return self.drive.backward(*args, **kwargs)

    def left(self, *args, **kwargs):
        if hasattr(self, "drive"):
            return self.drive.left(*args, **kwargs)

    def right(self, *args, **kwargs):
        if hasattr(self, "drive"):
            return self.drive.right(*args, **kwargs)

    def rotate(self, *args, **kwargs):
        if hasattr(self, "drive"):
            return self.drive.rotate(*args, **kwargs)

    def stop_rotation(self, *args, **kwargs):
        if hasattr(self, "drive"):
            return self.drive.stop_rotation(*args, **kwargs)

    def stop(self, *args, **kwargs):
        if hasattr(self, "drive"):
            return self.drive.stop(*args, **kwargs)

    def target_lock_drive_angle(self, *args, **kwargs):
        if hasattr(self, "drive"):
            return self.drive.target_lock_drive_angle(*args, **kwargs)

    def calibrate(self, *args, **kwargs):
        if hasattr(self, "pan_tilt"):
            return self.pan_tilt.calibrate(*args, **kwargs)

    @property
    def pan_servo(self):
        if hasattr(self, "pan_tilt"):
            return self.pan_tilt.pan_servo

    @property
    def tilt_servo(self):
        if hasattr(self, "pan_tilt"):
            return self.pan_tilt.tilt_servo



import configparser
from os.path import exists, isdir, join
from pathlib import Path


class TwoServoAssemblyCalibrator:
    CALIBRATION_FILE_DIR = ".config/pi-top/sdk"

    def __init__(self, filename, section_name, servo_lookup_dict):
        self.calibration_file_name = filename
        self.section_name = section_name
        self.servo_lookup_dict = servo_lookup_dict
        self.__calibration_file_path = join(
            str(Path.home()), self.CALIBRATION_FILE_DIR, self.calibration_file_name
        )

    def calibrate(self, save=True, reset=False):
        """Calibrates the assembly to work in optimal conditions. Based on the
        provided arguments, it will either load the calibration values stored
        in the pi-top, or it will run the calibration process, requesting the
        user input in an interactive fashion.

        :param bool reset:
            If `true`, the existing calibration values will be reset, and the calibration process will be started.
            If set to `false`, the calibration values will be retrieved from the calibration file.
        :param bool save:
            If `reset` is `true`, this parameter will cause the calibration values to be stored to the calibration file if set to `true`.
            If `save=False`, the calibration values will only be used for the current session.
        """
        if not reset and exists(self.__calibration_file_path):
            return self.__load_calibration()

        for servo_name, servo_obj in self.servo_lookup_dict.items():
            servo_obj.zero_point = 0
            value = self.__calibrate_servo(servo_name, servo_obj)
            if save and value is not None:
                self.__save_calibration(
                    section=self.section_name, values_dict={servo_name: value}
                )

        print("Calibration finished.")

    def __calibrate_servo(self, servo_name, servo_obj):
        print(f"Starting {servo_name} servo motor calibration.")

        while True:
            servo_obj.target_angle = 0
            print("Enter a angle to use as zero point.")
            user_zero_setting = input("Value: ")
            try:
                user_zero_setting = int(user_zero_setting)
                if user_zero_setting < int(
                    servo_obj.angle_range[0]
                ) or user_zero_setting > int(servo_obj.angle_range[1]):
                    print(
                        f"Invalid angle value {user_zero_setting}. "
                        f" Please enter a value in the range {servo_obj.angle_range}"
                    )
                    continue
            except ValueError:
                print("Please, enter a valid value")
                continue

            servo_obj.target_angle = user_zero_setting

            print("Is the servo horn aligned with the body?")
            finished_calibration = str(
                input("'y' to accept, 'n' to try again, 'e' to exit: ")
            )
            if finished_calibration.upper() == "Y":
                servo_obj.zero_point = user_zero_setting
                return user_zero_setting
            elif finished_calibration.upper() == "E":
                return None

    def __load_calibration(self):
        if exists(self.__calibration_file_path):
            config = configparser.ConfigParser()
            config.read(self.__calibration_file_path)

            if self.section_name in config.sections():
                section_config = config[self.section_name]
                for servo_name, servo_obj in self.servo_lookup_dict.items():
                    if section_config.get(servo_name):
                        servo_obj.zero_point = int(section_config.get(servo_name))

    def __save_calibration(self, section, values_dict):
        def create_config_file():
            conf_file_dir = join(str(Path.home()), self.CALIBRATION_FILE_DIR)
            if not isdir(conf_file_dir):
                Path(conf_file_dir).mkdir(parents=True, exist_ok=True)
            if not exists(self.__calibration_file_path):
                Path(self.__calibration_file_path).touch()

        if not exists(self.__calibration_file_path):
            create_config_file()

        config = configparser.ConfigParser()
        config.read(self.__calibration_file_path)
        if section not in config:
            config[section] = {}

        config[section].update({k: str(v) for k, v in values_dict.items()})
        with open(self.__calibration_file_path, "w") as configfile:
            config.write(configfile)



from typing import Callable, Optional, Tuple

_Limits = Tuple[Optional[float], Optional[float]]
_Components = Tuple[float, float, float]
_Tunings = Tuple[float, float, float]

def _clamp(value: Optional[float], limits: _Limits) -> Optional[float]: ...

_current_time: Callable[[], float]

class PID(object):
    Kp: float
    Ki: float
    Kd: float
    setpoint: float
    sample_time: Optional[float]
    output_limits: _Limits
    proportional_on_measurement: bool
    error_map: Optional[Callable[[float], float]]
    def __init__(
        self,
        Kp: float = ...,
        Ki: float = ...,
        Kd: float = ...,
        setpoint: float = ...,
        sample_time: Optional[float] = ...,
        output_limits: _Limits = ...,
        auto_mode: bool = ...,
        proportional_on_measurement: bool = ...,
        error_map: Optional[Callable[[float], float]] = ...,
    ) -> None: ...
    def __call__(self, input_: float, dt: Optional[float] = ...) -> Optional[float]: ...
    def __repr__(self) -> str: ...
    @property
    def components(self) -> _Components: ...
    @property
    def tunings(self) -> _Tunings: ...
    @tunings.setter
    def tunings(self, tunings: _Tunings) -> None: ...
    @property
    def auto_mode(self) -> bool: ...
    @auto_mode.setter
    def auto_mode(self, enabled: bool) -> None: ...
    def set_auto_mode(
        self, enabled: bool, last_output: Optional[float] = ...
    ) -> None: ...
    @property
    def output_limits(self) -> _Limits: ...
    @output_limits.setter
    def output_limits(self, limits: _Limits) -> None: ...
    def reset(self) -> None: ...



"""Pi-top Python SDK."""

import os
import sys

from setuptools import setup

if not sys.version_info >= (3, 7):
    raise ValueError("This package requires Python 3.7 or above")

HERE = os.path.abspath(os.path.dirname(__file__))

__version__ = os.environ.get("PYTHON_PACKAGE_VERSION", "0.0.1.dev1").replace('"', "")
assert __version__ != ""

__project__ = "pitop.robotics"
__author__ = "pi-top"
__author_email__ = "deb-maintainers@pi-top.com"

__url__ = "https://github.com/pi-top/pi-top-Python-SDK"
__platforms__ = "ALL"

# https://pypi.org/classifiers/
__classifiers__ = [
    "Development Status :: 5 - Production/Stable",
    "Intended Audience :: Education",
    "Intended Audience :: Developers",
    "Topic :: Education",
    "Topic :: System :: Hardware",
    "License :: OSI Approved :: Apache Software License",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: Implementation :: PyPy",
]

__keywords__ = [
    "pi-top",
    "raspberrypi",
    "battery",
]

__requires__ = [
    f"pitop.system=={__version__}",
    f"pitop.pma=={__version__}",
    f"pitop.core=={__version__}",
    "numpy>1.19.5,<2.0.0",
    "scipy>=1.6.0,<2.0.0",
]


def main():
    import io

    with io.open(os.path.join(HERE, "README.rst"), "r") as readme:
        setup(
            name=__project__,
            version=__version__,
            description=__doc__,
            long_description=readme.read(),
            classifiers=__classifiers__,
            author=__author__,
            author_email=__author_email__,
            url=__url__,
            license=[
                c.rsplit("::", 1)[1].strip()
                for c in __classifiers__
                if c.startswith("License ::")
            ][0],
            keywords=__keywords__,
            packages=["pitop.robotics"],
            include_package_data=True,
            platforms=__platforms__,
            install_requires=__requires__,
        )


if __name__ == "__main__":
    main()



from pitop.core.mixins import Recreatable, Stateful
from pitop.pma import ServoMotor, ServoMotorSetting
from pitop.robotics.two_servo_assembly_calibrator import TwoServoAssemblyCalibrator


class PincerController(Stateful, Recreatable):
    """Represents a pincer that uses two servo motors connected parallel to
    each other."""

    CALIBRATION_FILE_NAME = "pincers.conf"
    _left_pincer = None
    _right_pincer = None

    def __init__(self, left_pincer_port="S3", right_pincer_port="S0", name="pincers"):
        self.name = name
        self._left_pincer = ServoMotor(left_pincer_port)
        self._right_pincer = ServoMotor(right_pincer_port)
        self.__left_pincer_setting = ServoMotorSetting()
        self.__right_pincer_setting = ServoMotorSetting()

        Stateful.__init__(self, children=["_left_pincer", "_right_pincer"])
        Recreatable.__init__(
            self,
            config_dict={
                "left_pincer_port": left_pincer_port,
                "right_pincer_port": right_pincer_port,
                "name": self.name,
            },
        )

    def close(self, speed: int = 100, angle: int = 0):
        self.__left_pincer_setting.speed = speed
        self.__left_pincer_setting.angle = angle

        self.__right_pincer_setting.speed = speed
        self.__right_pincer_setting.angle = -angle

        self.pincer_move(
            left_servo_setting=self.__left_pincer_setting,
            right_servo_setting=self.__right_pincer_setting,
        )

    def open(self, speed: int = 50, angle: int = 45):
        self.__left_pincer_setting.speed = speed
        self.__left_pincer_setting.angle = -angle

        self.__right_pincer_setting.speed = speed
        self.__right_pincer_setting.angle = angle

        self.pincer_move(
            left_servo_setting=self.__left_pincer_setting,
            right_servo_setting=self.__right_pincer_setting,
        )

    def pincer_move(self, left_servo_setting, right_servo_setting):
        self._left_pincer.setting = left_servo_setting
        self._right_pincer.setting = right_servo_setting

    def calibrate(self, save=True, reset=False):
        """Calibrates the assembly to work in optimal conditions. Based on the
        provided arguments, it will either load the calibration values stored
        in the pi-top, or it will run the calibration process, requesting the
        user input in an interactive fashion.

        :param bool reset:
            If `true`, the existing calibration values will be reset, and the calibration process will be started.
            If set to `false`, the calibration values will be retrieved from the calibration file.
        :param bool save:
            If `reset` is `true`, this parameter will cause the calibration values to be stored to the calibration file if set to `true`.
            If `save=False`, the calibration values will only be used for the current session.
        """
        calibration_object = TwoServoAssemblyCalibrator(
            filename=self.CALIBRATION_FILE_NAME,
            section_name="PINCERS",
            servo_lookup_dict={
                "left_pincer_zero_point": self._left_pincer,
                "right_pincer_zero_point": self._right_pincer,
            },
        )
        calibration_object.calibrate(save, reset)



from time import time

from pitop.pma.servo_controller import ServoHardwareSpecs

from .simple_pid import PID


class PanTiltObjectTracker:
    _pid_tunings = {
        "slow": {"kp": 0.075, "ki": 0.002, "kd": 0.04},
        "normal": {"kp": 0.25, "ki": 0.005, "kd": 0.1},
    }
    _target_lock_range = 10
    _slow_fps_limit = 5.0

    def __init__(self, pan_servo, tilt_servo):
        self.__pan_servo = pan_servo
        self.__tilt_servo = tilt_servo
        self._previous_time = time()
        self.pan_pid = PID(
            setpoint=0,
            output_limits=(
                -ServoHardwareSpecs.SPEED_RANGE,
                ServoHardwareSpecs.SPEED_RANGE,
            ),
        )
        self.tilt_pid = PID(
            setpoint=0,
            output_limits=(
                -ServoHardwareSpecs.SPEED_RANGE,
                ServoHardwareSpecs.SPEED_RANGE,
            ),
        )
        self.__set_pid_tunings(pid_mode="normal")

    def __call__(self, center):
        current_time = time()
        dt = current_time - self._previous_time
        if dt > 1 / self._slow_fps_limit:
            pid_mode = "slow"
        else:
            pid_mode = "normal"
        self._previous_time = current_time
        self.__set_pid_tunings(pid_mode=pid_mode)

        x, y = center

        if abs(x) < self._target_lock_range:
            self.__pan_servo.sweep(speed=0)
            self.pan_pid.reset()
        else:
            pan_speed = self.pan_pid(x)
            self.__pan_servo.sweep(pan_speed)

        if abs(y) < self._target_lock_range:
            self.__tilt_servo.sweep(speed=0)
            self.tilt_pid.reset()
        else:
            tilt_speed = self.tilt_pid(y)
            self.__tilt_servo.sweep(tilt_speed)

    def __set_pid_tunings(self, pid_mode):
        self.pan_pid.tunings = list(self._pid_tunings[pid_mode].values())
        self.tilt_pid.tunings = list(self._pid_tunings[pid_mode].values())

    def reset(self):
        self.pan_pid.reset()
        self.tilt_pid.reset()

    def stop(self):
        self.__pan_servo.sweep(0)
        self.__tilt_servo.sweep(0)
        self.reset()



from pitop.robotics.drive_controller import DriveController
from pitop.system.pitop import Pitop


class BlockPiRover(Pitop):
    """A rover class for use with BlockPi coding in Further, with a simplified
    API.

    Inherits from Pitop. Constructor adds a DriveController to simplify
    use. DriveController methods are made available on the instance.
    """

    def __init__(self, left_motor="M3", right_motor="M0"):
        Pitop.__init__(self)

        drive = DriveController(
            left_motor_port=left_motor, right_motor_port=right_motor
        )
        self.add_component(drive)

    def forward(self, *args, **kwargs):
        self.drive.forward(*args, **kwargs)

    def backward(self, *args, **kwargs):
        self.drive.backward(*args, **kwargs)

    def left(self, *args, **kwargs):
        self.drive.left(*args, **kwargs)

    def right(self, *args, **kwargs):
        self.drive.right(*args, **kwargs)

    def stop(self, *args, **kwargs):
        self.drive.stop(*args, **kwargs)

    def move(self, *args, **kwargs):
        self.drive.robot_move(*args, **kwargs)
